//
// Cforall Version 1.0.0 Copyright (C) 2019 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// Time.cc --
//
// Author           : Thierry Delisle
// Created On       : Mon Mar 04 15:16:07 2019
// Last Modified By :
// Last Modified On :
// Update Count     :
//

#include "Time.h"

#include <cassert>
#include <chrono>
#include <cstdint>
#include <cstring>
#include <iostream>
#include <iomanip>
#include <stack>

namespace Stats {
	namespace Time {
#		if !defined(NO_TIME_STATISTICS)
			extern bool enabled;

			Base::TreeTop top;

			class TimerNode : public Base::Tree<top> {
			public:
				TimerNode(const char * const name )
					: Base::Tree<top>(name)
				{}

				TimerNode(const char * const name, Base::Tree<top> * parent)
					: Base::Tree<top>(name, parent)

				{}

				virtual void print(std::ostream & os) override {
					assert(finished);
					std::chrono::duration<double> diff = end - begin;
					os << diff.count();
				}

				void start() {
					begin = std::chrono::high_resolution_clock::now();
				}

				void finish() {
					end = std::chrono::high_resolution_clock::now();
					finished = true;
				}

			protected:
				virtual ~TimerNode() = default;

			private:
				bool finished = false;

				typedef  std::chrono::time_point<std::chrono::high_resolution_clock> point_t;
				point_t begin;
				point_t end;
			};

			std::stack<TimerNode *> nodes;

			void StartBlock(const char * const name) {
				if(!enabled) return;
				auto node = nodes.empty()
					? new TimerNode(name)
					: new TimerNode(name, nodes.top());

				nodes.push(node);
				node->start();
			}

			void StopBlock() {
				if(!enabled) return;
				nodes.top()->finish();
				nodes.pop();
			}

			void print() {
				if(!top.head) return;
				size_t nc = 0;
				Base::ForAll(top, 0, [&](Base::TreeImpl * node, size_t level) {
					nc = std::max(nc, (4 * level) + std::strlen(node->name));
				});

				const char * const title = "Timing Results";
				size_t nct = nc + 14;
				std::cerr << std::string(nct, '=') << std::endl;
				std::cerr << std::string((nct - std::strlen(title)) / 2, ' ');
				std::cerr << title << std::endl;
				std::cerr << std::string(nct, '-') << std::endl;


				Base::ForAll(top, 0, [&](Base::TreeImpl * node, size_t level) {
					std::cerr << std::string(level * 4, ' ');
					std::cerr << node->name;
					std::cerr << std::string(nc - ((level * 4) + std::strlen(node->name)), ' ');
					std::cerr << " | ";
					std::cerr << std::setw(9);
					node->print(std::cerr);
					std::cerr << " |";
					std::cerr << '\n';
				}, true);

				std::cerr << std::string(nct, '-') << std::endl;
			}
#		endif
	}
}